import sqlite3

class SQLiteManager(object):
    def __init__(self, dbname):
        self.connection = sqlite3.connect('%s.sqlite' % dbname)
        self.connection.row_factory = sqlite3.Row
        self.cursor = self.connection.cursor()

    def create_table(self, table_type, name_peak_list = ""):

        self.name_peak_list = name_peak_list

        if table_type == "single_peak_search":
            self.cursor.execute("DROP TABLE IF EXISTS single_peak_search_%s;" % (self.name_peak_list))
            self.cursor.execute("""
                CREATE TABLE single_peak_search_%s (
                compoundid char(6) NOT NULL,
                ion char(20) DEFAULT NULL,
                emp_formula text DEFAULT NULL,
                names text DEFAULT NULL,
                peak decimal(12,7) DEFAULT NULL,
                peak_intensity decimal(18,8) DEFAULT NULL,
                theo_mass_adduct decimal(12,7) DEFAULT NULL,
                ppm_error decimal(12,2) DEFAULT NULL,
                exact_mass decimal(12,7) DEFAULT NULL,
                ion_mode char(3) DEFAULT NULL,
                primary key  (compoundid, ion, peak, peak_intensity)
                );""" % (self.name_peak_list))


        if table_type == "transformation_mapping":
            self.cursor.execute("DROP TABLE IF EXISTS transformations_%s;" % (self.name_peak_list))
            self.cursor.execute("""
                CREATE TABLE transformations_%s (
                peaka decimal(12,7) DEFAULT NULL,
                peakb decimal(12,7) DEFAULT NULL,
                ion_peaka char(20) DEFAULT NULL,
                ion_peakb char(20) DEFAULT NULL,
                peaka_intensity decimal(18,8) DEFAULT NULL,
                peakb_intensity decimal(18,8) DEFAULT NULL,        
                transformation char(50) DEFAULT NULL,
                peak_pair_adducts_diff decimal(12,7) DEFAULT NULL,
                theo_peak_pair_adducts_diff decimal(12,7) DEFAULT NULL,
                ppm_error decimal(12,2) DEFAULT NULL,
                type int(2) DEFAULT NULL,
                ion_mode char(3) DEFAULT NULL
                );""" % (self.name_peak_list))


        if table_type == "compounds":

            self.cursor.execute("""DROP TABLE IF EXISTS compounds;""")
            self.cursor.execute("""DROP TABLE IF EXISTS maps_compounds;""")
            self.cursor.execute("""DROP TABLE IF EXISTS map_names;""")
            self.cursor.execute("""DROP TABLE IF EXISTS maps_formulae;""")

            self.cursor.execute("""
                CREATE TABLE compounds (
                compoundid char(6) not null,
                names text,
                emp_formula text,
                theo_mass decimal(12,7) default null,
                theo_mass_adduct decimal(12,7) default null,
                ion char(20),
                ion_mode char(3),
                PRIMARY KEY (compoundid, ion, ion_mode)
                );""")

            self.cursor.execute("""
                CREATE TABLE map_names (
                mapid char(8) not null,
                mapname tinytext,
                PRIMARY KEY (mapid)
                );""")

            self.cursor.execute("""
                CREATE TABLE maps_compounds (
                compoundid char(6) not null,
                mapid char(8) not null,
                PRIMARY KEY (mapid, compoundid)
                );""")

            self.cursor.execute("""
                CREATE TABLE maps_formulae (
                mapid char(8) not null,
                emp_formula text not null,
                PRIMARY KEY (mapid, emp_formula)
                );""")

        if table_type == "reactant_pairs":
            
            self.cursor.execute("drop TABLE if exists reactions;")
            self.cursor.execute("drop TABLE if exists reactant_pairs;")
                
            self.cursor.execute("""
                CREATE TABLE reactions (
                id mediumint(9) not null,
                reactionid char(6) not null,
                type char(1) default null,
                PRIMARY KEY (id, reactionid, type)
                );""")

            self.cursor.execute("""
                CREATE TABLE reactant_pairs (
                id int(11) default null,
                compoundid_1 char(6) not null,
                compoundid_2 char(6) not null,
                ion_compoundid_1 char(20) default null,
                ion_compoundid_2 char(20) default null,
                type int(1) default null,
                peak_pair_diff decimal(12,7) default null,
                peak_pair_adducts_diff decimal(12,7) default null,
                transformation char(50) default null,
                ion_mode char(3) default null,
                PRIMARY KEY  (compoundid_1, compoundid_2, ion_compoundid_1, ion_compoundid_2, type, ion_mode)
                );""")

        if table_type == "temp_reactant_pairs":
            self.cursor.execute("DROP TABLE IF EXISTS temp_reactant_pairs;")
            self.cursor.execute("""
                    CREATE TEMPORARY TABLE temp_reactant_pairs (
                      id1 int(11) not null,
                      compoundid_1 char(6) not null,
                      id2 int(11) not null,
                      compoundid_2 char(6) not null
                    );
                    """)
            self.cursor.execute("""INSERT INTO temp_reactant_pairs SELECT DISTINCT reactant_pairs.id AS id1, reactant_pairs.compoundid_1, cs.id AS id2, cs.compoundid_2
                FROM reactant_pairs inner join 
                reactant_pairs AS cs on reactant_pairs.compoundid_2 = cs.compoundid_1 and reactant_pairs.compoundid_1 != cs.compoundid_2;""")
            self.save()

        if table_type == "unique_transformations":
            self.cursor.execute("DROP TABLE IF EXISTS unique_transformations;")

            self.cursor.execute("""CREATE TABLE unique_transformations AS SELECT DISTINCT
                ion_compoundid_1, ion_compoundid_2, transformation, peak_pair_adducts_diff, type, ion_mode
                FROM reactant_pairs ORDER BY peak_pair_adducts_diff;""")
            
            self.cursor.execute("""CREATE UNIQUE INDEX PK_unique_transformations ON unique_transformations
                (ion_compoundid_1, ion_compoundid_2, transformation, peak_pair_adducts_diff, type, ion_mode);""")


        if table_type == "reactant_pairs_peaklist":
            self.cursor.execute("DROP TABLE IF EXISTS reactant_pairs_temp_%s;" % (self.name_peak_list))
            self.cursor.execute("""CREATE TEMPORARY TABLE reactant_pairs_temp_%s (
                compoundid_1 char(6) DEFAULT NULL,
                ion_peaka char(20) DEFAULT NULL,
                compoundid_2 char(6) DEFAULT NULL,
                ion_peakb char(20) DEFAULT NULL,
                transformation char(50) DEFAULT NULL,
                ppm_error decimal(12,2) DEFAULT NULL,        
                peaka decimal(12,7) DEFAULT NULL,
                peakb decimal(12,7) DEFAULT NULL,
                peaka_intensity decimal(18,8) DEFAULT NULL,
                peakb_intensity decimal(18,8) DEFAULT NULL, 
                peak_pair_adducts_diff decimal(12,7) DEFAULT NULL,
                type int(1) DEFAULT NULL,
                ion_mode char(3) DEFAULT NULL
                );""" % (self.name_peak_list))

            self.cursor.execute("DROP TABLE IF EXISTS reactant_pairs_%s;" % (self.name_peak_list))
            self.cursor.execute("""CREATE TABLE reactant_pairs_%s (
                compoundid_1 char(6) DEFAULT NULL,
                ion_peaka char(20) DEFAULT NULL,
                compoundid_2 char(6) DEFAULT NULL,
                ion_peakb char(20) DEFAULT NULL,
                transformation char(50) DEFAULT NULL,
                ppm_error decimal(12,2) DEFAULT NULL,        
                peaka decimal(12,7) DEFAULT NULL,
                peakb decimal(12,7) DEFAULT NULL,
                peaka_intensity decimal(18,8) DEFAULT NULL,
                peakb_intensity decimal(18,8) DEFAULT NULL,
                peak_pair_adducts_diff decimal(12,7) DEFAULT NULL,
                type int(1) DEFAULT NULL,
                ion_mode char(3) DEFAULT NULL
                );""" % (self.name_peak_list))

            self.cursor.execute("DROP TABLE IF EXISTS compounds_from_reactant_pairs_%s;" % (self.name_peak_list))
            self.cursor.execute("""CREATE TABLE compounds_from_reactant_pairs_%s (
                compoundid char(6) NOT NULL,
                ion char(20) DEFAULT NULL,
                emp_formula text DEFAULT NULL,
                names text DEFAULT NULL,
                peak decimal(12,7) DEFAULT NULL,
                peak_intensity decimal(18,8) DEFAULT NULL,
                theo_mass_adduct decimal(12,7) DEFAULT NULL,
                ppm_error decimal(12,2) DEFAULT NULL,
                exact_mass decimal(12,7) DEFAULT NULL,
                type int(1) DEFAULT NULL,
                ion_mode char(3) DEFAULT NULL,
                primary key  (compoundid, ion, peak, type)
                );""" % (self.name_peak_list))

        if table_type == "maps":
            self.cursor.execute("DROP TABLE IF EXISTS maps__compounds__%s;" % (self.name_peak_list))
            self.cursor.execute("""
                CREATE TABLE maps__compounds__%s (
                mapid char(8) not null,
                compoundid char(6) NOT NULL,
                transformation_type int(1) NOT NULL,
                primary key  (compoundid, mapid, transformation_type)
                );""" % (self.name_peak_list))

            self.cursor.execute("DROP TABLE IF EXISTS maps_coverage__compounds__%s;" % (self.name_peak_list))
            self.cursor.execute("""
                CREATE TABLE maps_coverage__compounds__%s (
                mapid char(8) not null,
                transformation_type int(1) not null,
                mapname text not null,
                url text NOT NULL,
                count int(11) NOT NULL,
                total int(11) NOT NULL,
                coverage int(11) NOT NULL,
                primary key (mapid, transformation_type)
                );""" % (self.name_peak_list))

            self.cursor.execute("DROP TABLE IF EXISTS maps__formulae__%s;" % (self.name_peak_list))
            self.cursor.execute("""
                CREATE TABLE maps__formulae__%s (
                mapid char(8) not null,
                emp_formula text NOT NULL,
                primary key  (emp_formula, mapid)
                );""" % (self.name_peak_list))

            self.cursor.execute("DROP TABLE IF EXISTS maps_coverage__formulae__%s;" % (self.name_peak_list))
            self.cursor.execute("""
                CREATE TABLE maps_coverage__formulae__%s (
                mapid char(8) not null,
                transformation_type int(1) not null,
                mapname text not null,
                count int(11) NOT NULL,
                total int(11) NOT NULL,
                coverage int(11) NOT NULL,
                primary key (mapid, transformation_type)
                );""" % (self.name_peak_list))

        if table_type == "formulae":
            self.cursor.execute("DROP TABLE IF EXISTS formulae_%s;" % (self.name_peak_list))
            self.cursor.execute("""CREATE TABLE formulae_%s (
                peak decimal(12,7) DEFAULT NULL,
                intensity decimal(16,8) DEFAULT NULL,
                theoretical_mass decimal(12,7) DEFAULT NULL,
                ppm_error decimal(12,7) DEFAULT NULL,
                formula text DEFAULT NULL,
                ion char(15) DEFAULT NULL,
                lewis int(1) DEFAULT NULL,
                senior int(1) DEFAULT NULL,
                HC int(1) DEFAULT NULL,
                NOPSC int(1) DEFAULT NULL,
                ion_mode char(3) DEFAULT NULL
                );""" % (self.name_peak_list))

        if table_type == "peak_patterns":
            self.cursor.execute("DROP TABLE IF EXISTS peak_patterns_%s;" % (self.name_peak_list))
            self.cursor.execute("""
                CREATE TABLE peak_patterns_%s (
                id int(11) DEFAULT NULL,
                peak_pair_diff decimal(12,7) DEFAULT NULL,
                peaka decimal(12,7) DEFAULT NULL,
                peakb decimal(12,7) DEFAULT NULL,
                peaka_intensity decimal(18,8) DEFAULT NULL,
                peakb_intensity decimal(18,8) DEFAULT NULL,
                peaka_label char(15) DEFAULT NULL,
                peakb_label char(15) DEFAULT NULL,
                pattern_type text DEFAULT NULL,
                num_peaks int(2) DEFAULT NULL,
                carbons int(3) DEFAULT NULL,
                rules int(1) DEFAULT NULL,
                ion_mode char(3) DEFAULT NULL,
                PRIMARY KEY (peaka, peakb, peaka_label, peakb_label)
                );""" % (self.name_peak_list))



    def execute_complex(self, query_type, input_list, name_peak_list = ""):

        if name_peak_list != "":
            self.name_peak_list = name_peak_list

        if query_type == "reactant_pairs_temp":

            self.cursor.execute("""
            INSERT INTO reactant_pairs_temp_%s
            SELECT reactant_pairs.compoundid_1, reactant_pairs.ion_compoundid_1, reactant_pairs.compoundid_2, reactant_pairs.ion_compoundid_2,
            reactant_pairs.transformation, ppm_error, peaka, peakb, peaka_intensity, peakb_intensity, reactant_pairs.peak_pair_adducts_diff,
            reactant_pairs.type, reactant_pairs.ion_mode

            FROM reactant_pairs INNER JOIN transformations_%s as diffs WHERE
        
            reactant_pairs.compoundid_1 = '%s' AND diffs.peaka = %s
            AND reactant_pairs.ion_compoundid_1 = '%s'
            AND reactant_pairs.transformation = diffs.transformation
            AND reactant_pairs.type <= %s AND reactant_pairs.type = diffs.type
            AND reactant_pairs.ion_mode = '%s' AND reactant_pairs.ion_mode = diffs.ion_mode;

            """ % tuple(input_list))
            
        if query_type == "reactant_pairs":
            self.save()
            self.cursor.execute("""INSERT INTO reactant_pairs_%s SELECT DISTINCT * from reactant_pairs_temp_%s;""" % (self.name_peak_list, self.name_peak_list))
            self.save()
            self.cursor.execute("""INSERT INTO compounds_from_reactant_pairs_%s
                (compoundid, ion, emp_formula, names, peak, peak_intensity, theo_mass_adduct, ppm_error, exact_mass, type, ion_mode)
                SELECT distinct compoundid_1, ion_peaka, emp_formula, names, peaka, peaka_intensity, theo_mass_adduct, 
                (peaka-theo_mass_adduct)/(peaka*0.000001), theo_mass, type, reactant_pairs_%s.ion_mode
                FROM reactant_pairs_%s
                INNER JOIN compounds WHERE
                compounds.compoundid = reactant_pairs_%s.compoundid_1 AND compounds.ion = reactant_pairs_%s.ion_peaka
                ;""" % (self.name_peak_list, self.name_peak_list, self.name_peak_list, self.name_peak_list, self.name_peak_list))
            
        if query_type == "reactant_pairs_maps":
            self.cursor.execute("""INSERT INTO maps__compounds__%s (mapid, compoundid, transformation_type)
                SELECT distinct maps_compounds.mapid, compounds_from_reactant_pairs_%s.compoundid, compounds_from_reactant_pairs_%s.type
                FROM compounds_from_reactant_pairs_%s
                INNER JOIN maps_compounds WHERE
                compounds_from_reactant_pairs_%s.compoundid = maps_compounds.compoundid
                ;""" % (input_list[0], self.name_peak_list, self.name_peak_list, self.name_peak_list, self.name_peak_list))

            self.cursor.execute("""INSERT INTO maps__formulae__%s (mapid, emp_formula)
                SELECT distinct maps_formulae.mapid, compounds_from_reactant_pairs_%s.emp_formula
                FROM compounds_from_reactant_pairs_%s
                INNER JOIN maps_formulae WHERE
                compounds_from_reactant_pairs_%s.emp_formula = maps_formulae.emp_formula
                ;""" % (input_list[0], self.name_peak_list, self.name_peak_list, self.name_peak_list))
            self.save()

        if query_type == "single_peak_search_maps":
            self.cursor.execute("""INSERT INTO maps__compounds__%s (mapid, compoundid, transformation_type)
                SELECT distinct maps_compounds.mapid, single_peak_search_%s.compoundid, 0
                FROM single_peak_search_%s
                INNER JOIN maps_compounds WHERE
                single_peak_search_%s.compoundid = maps_compounds.compoundid
                ;""" % (input_list[0], self.name_peak_list, self.name_peak_list, self.name_peak_list))

            self.cursor.execute("""INSERT INTO maps__formulae__%s (mapid, emp_formula)
                SELECT distinct maps_formulae.mapid, single_peak_search_%s.emp_formula
                FROM single_peak_search_%s
                INNER JOIN maps_formulae WHERE
                single_peak_search_%s.emp_formula = maps_formulae.emp_formula
                ;""" % (input_list[0], self.name_peak_list, self.name_peak_list, self.name_peak_list))
            self.save()
            
    def index(self, index_name, table_name, columns):

        if len(columns) > 1:
            columns = ", ".join(columns)
        else:
            columns = columns[0]

        self.cursor.execute("""
        CREATE INDEX idx_%s
        on %s (%s);
        """ % (index_name, table_name, columns))

    def attach(self, ion_mode, db_name_to_attach):
        self.cursor.execute("attach 'MI_DB_%s.sqlite' as '%s'" % (ion_mode, db_name_to_attach))
       
    def drop(self, table_name):
        self.cursor.execute("DROP TABLE IF EXISTS %s;" % (table_name))
       
    def __len__(self):
        a = 0
        for i in self.__iter__():
            a += 1
        return a
       
    def execute(self, command):
        self.cursor.execute(command)
                  
    def max(self, table_name, column, where=""):
        self.cursor.execute("SELECT MAX(%s) FROM %s %s;" % (column, table_name, where))
        return self.cursor.fetchone()[0]

    def min(self, table_name, column, where=""):
        self.cursor.execute("SELECT MIN(%s) FROM %s %s;" % (column, table_name, where))
        return self.cursor.fetchone()[0]

    def count(self, table_name, column, where=""):
        self.cursor.execute("SELECT count(%s) FROM %s %s;" % (column, table_name, where))
        return self.cursor.fetchone()[0]
       
    def select(self, table_name, columns, where=""):
        if len(columns) > 1:
            columns = ", ".join(map(str,columns))
        else:
               columns = columns[0]
        self.cursor.execute("SELECT %s FROM %s %s" % (columns, table_name, where))
        return self.cursor.fetchall()
                   
    def insert(self, table_name, columns_values):
        self.cursor.execute("INSERT INTO %s (%s) VALUES (%s)" % (table_name, ", ".join(map(str,columns_values.keys())), str(columns_values.values())[1 : -1]))

    def delete(self, table_name, where=""):
        self.cursor.execute("DELETE FROM %s %s" % (table_name, where))

    def insert_select(self, table_name_insert, columns_insert, columns_select, table_name_select, where=""):
        if len(columns_insert) > 1:
            columns_insert = ", ".join(map(str,columns_insert))
        else:
            columns_insert = columns_insert[0]

        if len(columns_select) > 1:
            columns_select = ", ".join(map(str,columns_select))
        else:
            columns_select = columns[0]

        self.cursor.execute("""INSERT INTO %s (%s)
        SELECT %s FROM %s %s;
        """ % (table_name_insert, columns_insert, columns_select, table_name_select, where))
           
    def fetchall(self, values_only=""):
        if values_only != "":
            trueValues = []
            for value in self.cursor.fetchall():
                if len(value) > 1:
                    trueValues.append(value)
                else:
                    trueValues.append(value[0])
            return trueValues
        else:
            return self.cursor.fetchall()
       
    def fetchone(self):
        return self.cursor.fetchone()[0]

    def save(self):
        self.connection.commit()
        
    def close(self):
        self.connection.close()

