#!/usr/local/bin/python

from SqliteManager import *

class Atoms:
    def __init__(self):
        self.lib = {"H":1.007825, "C":12.000000, "N":14.003074, "O":15.994915, "P":30.973763,
            "S":31.972072, "F":18.998403, "B":11.009305, "K":38.963708, "Na":22.989770,
            "Cl":34.968853, "Si":27.976928, "Se":79.916521, "Mg":23.985045, "Br":78.9183,
            "I":126.904477, "Ca":39.9625, "Ba":137.905236, "Fe": 55.934939, "R":0.000000}

        self.limits = {
            "C":range(0, 35),
            "H":range(0, 73),
            "N":range(0, 16),
            "O":range(0, 20),
            "P":range(0, 8),
            "S":range(0, 9)}
        self.atom_bonds = {'C':4, 'H':1, 'N':3, 'O':2, 'P':3, 'S':2}

    def calculate_limits(self, mass):
        for atom in self.limits:
            min_atoms = min([int(self.limits[atom][len(self.limits[atom])-1]), int(mass/self.lib[atom]) + 1])
            if  min_atoms <= 0:
                self.limits[atom] = [0]
            else:
                self.limits[atom] = range(self.limits[atom][0], min_atoms)
        return self.limits
        
    def alter_mass(self, name, mass):
        if type(mass) != float:
            print "mass: wrong format"
        else:
            if name in self.lib.keys():
                self.lib[name] = mass
            else:
                print "Name ion not in library"

    def alter_name(self, name, name_new):
        if name in self.lib.keys():
            mass = self.lib[name]
            del self.lib[name]
            self.lib[name_new] = mass
        else:
            print "Name ion not in library"

    def alter_limits(self, name, min_max):
        if name in self.limits.keys():
            self.limits[name] = range(min_max[0], min_max[1]+1)
        else:
            print "Name ion not in library"

    def add(self, name, mass):
        self.lib[name] = mass
        print "Do not forget to add atom limits"

    def add_limit(self, name, min, max):
        self.limits[name] = range(min, max+1)

    def delete(self, name):
        del self.lib[name]
        del self.limits[name]

    def __str__(self):
        print "Atoms in library:"
        for key in self.lib:
            print "%s\t%s" % (key, self.lib[key])
        print
        print "Atom limits:"
        for key in self.limits:
            print "%s\t%s-%s" % (key, min(self.limits[key]), max(self.limits[key]))
        print
        return ""

    def sort_abc_length(self, items = []):
        char, chars = [], []
        if items == []:
            items = self.lib.keys()
        for atom in items:
            if len(atom) == 1:
                char.append(atom)
            else:
                chars.append(atom)
        char.sort()
        chars.sort()
        self.sort_atoms_abc_length = char + chars
        return self.sort_atoms_abc_length

class Ions:    
    def __init__(self, ion_mode):
        self.ion_mode = ion_mode
        self.e = 0.0005486
        if self.ion_mode == "POS":
            self.lib = {"H+":1.007825-self.e, "Na+":22.989770-self.e, "K(39)+":38.963708-self.e}
            
        elif self.ion_mode == "NEG":
            self.lib = {"-H+": -(1.007825-self.e), "Cl(35)-": 34.968853+self.e, "Cl(37)-": 36.965903+self.e}

    def alter_mass(self, name, mass):
        if type(mass) != float:
            print "mass: wrong format"
        else:
            if name in self.lib.keys():
                self.lib[name] = mass
            else:
                print "Name ion not in library"

    def alter_name(self, name, name_new):
        if name in self.lib.keys():
            mass = self.lib[name]
            del self.lib[name]
            self.lib[name_new] = mass
        else:
            print "Name ion not in library"

    def add(self, name, mass):
        self.lib[name] = mass
            
    def delete(self, name):
        del self.lib[name]
    
    def __str__(self):
        print "Ions in library:"
        for key in self.lib:
            print "%s\t%s" % (key, self.lib[key])
        print
        return ""

    def sort_abc_length(self, items = []):
        if items == []:
            items = self.lib.keys()
        char, chars = [], []
        for ion in items:
            if len(ion) == 1:
                char.append(ion)
            else:
                chars.append(ion)
        char.sort()
        chars.sort()
        self.sort_ions_abc_length = char + chars
        return self.sort_ions_abc_length


class Formula:
    
    def __init__(self, formula, atoms, ions):
        
        self.atoms = atoms
        self.ions = ions
        self.formula = formula
        
        self.count_atoms = {}
        self.count_ions = {}

        self.theo_mass = 0.0
        self.theo_mass_adduct = 0.0

        self.formula_print_it = ""
        
        self.count()
        self.mass()
        self.lewis_senior()
        self.HNOPS()

    def count(self, debug=0):

        import re
        
        self.temp_formula = self.formula
        
        for atom in self.atoms.lib:
            self.count_atoms[atom] = 0
        for ion in self.ions.lib:
            self.count_ions[ion] = 0

        l_ions = self.ions.sort_abc_length()
        l_ions.reverse()

        for ion in l_ions:
            if ion in self.formula:
                self.temp_formula =  self.temp_formula.replace(ion,"")
                self.count_ions[ion] += 1

        l_atoms = self.atoms.sort_abc_length()
        l_atoms.reverse()

        for atom in l_atoms:
            str_pattern = []
            if "(" in atom:
                num = len(atom)
                atom_search = atom.replace("(","\D").replace(")","\D")
                str_pattern.append("%s\d*" % (atom_search))                
                pattern_char = re.compile(r'(%s)' % ("|".join(map(str,str_pattern))))
                result = pattern_char.findall(self.temp_formula)
                for item in result:
                    self.temp_formula = self.temp_formula.replace(item, "")
                    count = item.replace(atom, "")
                    if len(count) == 0:
                        self.count_atoms[atom] += 1
                    else:
                        self.count_atoms[atom] += int(count)
                 
            else:
                num = len(atom)
                str_pattern.append("%s\d*" % (atom))                
                pattern_char = re.compile(r'(%s)' % ("|".join(map(str,str_pattern))))
                result = pattern_char.findall(self.temp_formula)
                for item in result:
                    self.temp_formula = self.temp_formula.replace(item, "")
                    count = item.replace(atom, "")
                    if len(count) == 0:
                        self.count_atoms[atom] += 1
                    else:
                        self.count_atoms[atom] += int(count)
                        
        if len(self.temp_formula) > 0 and debug == 1:
            print "Wrong Format Before:", self.formula, "After:", self.temp_formula
            return

        return self.count_atoms, self.count_ions
    
    def mass(self):

        self.theo_mass = 0.0
        self.theo_mass_adduct = 0.0
        
        for atom in self.count_atoms:
            if self.count_atoms[atom] != 0:
                self.theo_mass += (self.count_atoms[atom] * (self.atoms.lib[atom]))
                
        self.theo_mass_adduct += self.theo_mass
        for ion in self.count_ions:
            if self.count_ions[ion] != 0:
                self.theo_mass_adduct += (self.count_ions[ion] * (self.ions.lib[ion]))

        return self.theo_mass, self.theo_mass_adduct
    
    def HNOPS(self):
        self.HNOPS_rule = {"HC":0, "NOPSC":0}
        if self.count_atoms['C'] == 0 or self.count_atoms['H'] == 0:
            self.HNOPS_rule["HC"]= 0
        elif self.count_atoms['C'] == 0 and self.count_atoms['H'] == 0:
            self.HNOPS_rule["HC"]= 0
        elif self.count_atoms['C'] != 0 and self.count_atoms['H'] != 0:
            if float(self.count_atoms['H'])/float(self.count_atoms['C']) > 0 and float(self.count_atoms['H']/self.count_atoms['C']) < 6:
                self.HNOPS_rule["HC"]= 1
            if float(self.count_atoms['H'])/float(self.count_atoms['C']) >= 6:
                self.HNOPS_rule["HC"]= 0

        #NOPS
        NOPS_check = []
        for atom in ['N','O','P','S']:
            try:
                check = float(float(self.count_atoms[atom]))/float(self.count_atoms['C'])
                NOPS_check.append(check)
            except ZeroDivisionError:
                NOPS_check.append(float(0))
        if NOPS_check[0] >= float(0) and NOPS_check[0] <= float(4) and NOPS_check[1] >= float(0) and NOPS_check[1] <= float(3) and NOPS_check[2] >= float(0) and NOPS_check[2] <= float(2) and NOPS_check[3] >= float(0) and NOPS_check[3] <= float(3):
            self.HNOPS_rule["NOPSC"] = 1
        if NOPS_check[0] > float(4) or NOPS_check[1] > float(3) or NOPS_check[2] > float(2) or NOPS_check[3] > float(3):
            self.HNOPS_rule["NOPSC"] = 0
        #print self.HNOPS_rule
        return self.HNOPS_rule
        
    def lewis_senior(self):

        self.lewis_senior_rule = {"lewis":0, "senior":""}
        
        atoms_sum = sum(self.count_atoms.values())
        lewis_sum = 0

        for atom in self.atoms.atom_bonds:
            lewis_sum += self.atoms.atom_bonds[atom] * self.count_atoms[atom]
        if lewis_sum%2 == 0:
            self.lewis_senior_rule["lewis"] = 1
        if lewis_sum%2 != 0:
            self.lewis_senior_rule["lewis"] = 0
        if lewis_sum >= ((atoms_sum-1)*2):
            self.lewis_senior_rule["senior"] = 1
        if lewis_sum < ((atoms_sum-1)*2):
            self.lewis_senior_rule["senior"] = 0
        return self.lewis_senior_rule

    def print_it(self, atom_count="", ion_count=""):

        if len(atom_count) > 0:
            self.count_atoms = atom_count
        if len(ion_count) > 0:
            self.count_ions = ion_count
        
        self.formula_print_it = ""

        for atom in self.atoms.sort_abc_length():
            try:
                if abs(self.count_atoms[atom]) > 1:
                    self.formula_print_it += atom + str(self.count_atoms[atom])
                if abs(self.count_atoms[atom]) == 1:
                    self.formula_print_it += atom
            except:
                pass

        for ion in self.ions.sort_abc_length():
            try:
                if self.count_ions[ion] > 1:
                    self.formula_print_it += ion + str(self.count_ions[ion])
                if self.count_ions[ion] == 1:
                    self.formula_print_it += ion
            except:
                pass

        return self.formula_print_it


    def diff(self, inp_formula):

        self.inp_formula = Formula(inp_formula, self.atoms, self.ions)
         
        self.formulae_diff = ""

        for item in self.count_atoms.keys():
             self.inp_formula.count_atoms[item] -= self.count_atoms[item]        

        for atom in self.atoms.sort_abc_length():
            if  self.inp_formula.count_atoms[atom] == 1:
                self.formulae_diff += "-" + str(atom)

            if  self.inp_formula.count_atoms[atom] == -1:
                self.formulae_diff += str(atom)
                
            if  self.inp_formula.count_atoms[atom] < -1:
                self.formulae_diff += str(atom) + str(abs(self.inp_formula.count_atoms[atom]))

            if  self.inp_formula.count_atoms[atom] > 1:
                self.formulae_diff += "-" + str(atom) + str(self.inp_formula.count_atoms[atom])

        return self.formulae_diff

class Calculate:
    
    def __init__(self):
        self.mass = 0.0
        self.ppm = 1.0
        self.theo_mass = 0.0
    
    def tolerance_peak(self, mass, ppm):
        self.mass = mass
        self.ppm = ppm
        self.min_tol = mass - (mass * 0.000001 * ppm)
        self.max_tol = mass + (mass * 0.000001 * ppm)
        return self.min_tol, self.max_tol

    def abs_error(self, mass, ppm):
        self.mass = mass
        self.ppm = ppm
        self.error = (mass * 0.000001 * ppm)
        return self.error
        
    def ppm_error(self, mass, theo_mass):
        return float(theo_mass-mass) / (theo_mass * 0.000001)

    def load_surface(self, fname_surface):
        try:
            r("library(GenKern)")
        except:
            r.install_packages("GenKern")
            r("library(GenKern)")
        r('load(file = "%s")' % (fname_surface))

    def mass_error_peak_pair_diff(self, avr_mrange, peak_pair_diff, name_surface):
        set_default_mode(BASIC_CONVERSION)
        r("mrange = %s" % (avr_mrange))
        r("peak_pair_diff = %s" % (peak_pair_diff))
        x = r("index_mrange = nearest(%s$x, mrange, outside=TRUE)" % (name_surface))
        y = r("index_peak_pair_diff = nearest(%s$y, peak_pair_diff, outside=TRUE)" % (name_surface))
        out = r("%s$z[index_mrange, index_peak_pair_diff]" % (name_surface))
        set_default_mode(NO_CONVERSION)
        return out

    def formulae(self, mass, ppm, atoms, ions):
        import copy
        self.tolerance_peak(mass, ppm)

        f = Formula("", atoms, ions)

        for ion in ions.lib:
            copy_atoms = copy.deepcopy(atoms)
            copy_atoms.calculate_limits(mass)

            for atom in copy_atoms.limits:
                for c in range(len(copy_atoms.limits[atom])):
                    copy_atoms.limits[atom][c] = c * copy_atoms.lib[atom]
            hit = 0
            wheels = map(iter, copy_atoms.limits.values())
            digits = []
            digits = [it.next( ) for it in wheels]

            while True:
                if (sum(digits) + ions.lib[ion])  >= self.min_tol and (sum(digits) + ions.lib[ion]) <= self.max_tol:
                    count = {}
                    hit = 1
                    for atom in range(len(digits)):
                        count[copy_atoms.limits.keys()[atom]] = int(digits[atom] / copy_atoms.lib[copy_atoms.limits.keys()[atom]])
                    yield {"mz":mass, "theoretical_mass":sum(digits) + ions.lib[ion],
                           "ppm_error":(mass-(sum(digits) + ions.lib[ion]))/((sum(digits) + ions.lib[ion])*0.000001),
                           "formula":f.print_it(count), "ion":ion, "HC":f.HNOPS()["HC"], "NOPSC":f.HNOPS()["NOPSC"],
                           "lewis":f.lewis_senior()["lewis"], "senior":f.lewis_senior()["senior"]}
                    
                for i in range(len(digits)-1, -1, -1):
                    try:
                        if sum(digits) > self.max_tol or hit == 1:
                            hit = 0
                            raise StopIteration
                        else:
                            digits[i] = wheels[i].next( )
                        break
                    except StopIteration:
                        wheels[i] = iter(copy_atoms.limits.values()[i])
                        digits[i] = wheels[i].next( )
                else:
                    break        


class PeakList:
    
    def __init__(self, file_peaklist):
        """read peaklist
        M/Z\tINTENSITY
        76.03931\t45005.555
        82.02633\t12778.42111
        83.02158\t109056.7006
        101.00316\t27555.51278
        """
        self.file_peaklist = file_peaklist
        self.intensities = {}
        self.mzs = {}
        
        f = open(self.file_peaklist)
        f.readline()
        for line in f:

            self.intensities[float(line.replace("\n","").split("\t")[0])] = float(line.replace("\n","").split("\t")[1])
            self.mzs[float(line.replace("\n","").split("\t")[1])] = float(line.replace("\n","").split("\t")[0])

    def list_mzs(self):
        return self.intensities.keys()

    def list_intensities(self):
        return self.mzs.keys()

    def intensity(self, mz_inp):
        return self.intensities[mz_inp]

    def mz(self, intensity_inp):
        return self.mzs[intensity_inp]

    def peak_pair_diffs(self, n = 2, peaks=[]):
        if len(peaks) == 0: 
            peaks = self.intensities.keys()
        if n == 0:
            yield []
        else:
            for i in xrange(len(peaks)):
                for cc in self.peak_pair_diffs(n-1, peaks[i+0:]):
                    if len([peaks[i]]+cc) == 2:

                        if peaks[i] < cc[0]:
                            self.peaka_intensity = self.intensity(peaks[i])
                            self.peakb_intensity = self.intensity(cc[0])
                            self.mean_peakab = (peaks[i] + cc[0])/2

                            yield {"peaka":peaks[i], "peaka_intensity":self.peaka_intensity, "peakb":cc[0], 
                               "peakb_intensity":self.peakb_intensity, "peak_pair_diff":cc[0] - peaks[i], "mean_peakab":self.mean_peakab}
                            
                        elif peaks[i] > cc[0] or peaks[i] == cc[0]:
                            self.peaka_intensity = self.intensity(peaks[i])
                            self.peakb_intensity = self.intensity(cc[0])
                            self.mean_peakab = (peaks[i] + cc[0])/2
                            
                            yield {"peaka":peaks[i], "peaka_intensity":self.peaka_intensity, "peakb":cc[0], 
                               "peakb_intensity":self.peakb_intensity, "peak_pair_diff":peaks[i] - cc[0], "mean_peakab":self.mean_peakab}

                    else:
                        yield [peaks[i]]+cc
                

class Identification:

    def __init__(self, file_name_peak_list, project_name, ion_mode, ppm):

        self.file_name_peak_list = file_name_peak_list
        self.name_peak_list = file_name_peak_list.split(".")[0]
        self.peak_list = PeakList(file_name_peak_list)

        self.ion_mode = ion_mode
        self.ppm = ppm

        self.database = SQLiteManager(project_name)
        self.database.attach(ion_mode, project_name)

        self.calc = Calculate()
        self.num_peaks = len(self.peak_list.intensities)

    def single_peak_search(self):
        self.database.create_table("single_peak_search", self.name_peak_list)
        peaks_done, perc = 0, 0.0
        print
        print "Single-peak search - %s" % (self.file_name_peak_list)
        print "------------------------------------------------------------------"
        peaks_done, perc = 0, 0.0
        self.num_peaks = len(self.peak_list.intensities)

        for mz in self.peak_list.intensities:
            
            self.calc.tolerance_peak(mz, self.ppm)
            
            self.database.insert_select("single_peak_search_%s" % (self.name_peak_list),

            ["compoundid", "ion", "emp_formula", "names", "peak", "peak_intensity",
            "theo_mass_adduct", "ppm_error", "exact_mass", "ion_mode"],

            ["compoundid", "ion", "emp_formula", "names", mz, self.peak_list.intensities[mz],
            "theo_mass_adduct", "(%s-theo_mass_adduct)/(theo_mass_adduct*0.000001)" % (mz), "theo_mass", "ion_mode"],
            "compounds",
            "WHERE theo_mass_adduct >= %s AND theo_mass_adduct <= %s" % (self.calc.min_tol, self.calc.max_tol))

            peaks_done += 1
            if float(peaks_done)/self.num_peaks >= perc:
                str_perc = str(perc * 100.0) + "%"
                perc += 0.1
                print str_perc,
        print
        print "------------------------------------------------------------------"
        print "DONE"
        print
        
        self.database.save()
        self.database.index("single_peak_search_%s" % (self.name_peak_list), "single_peak_search_%s" % (self.name_peak_list), ["peak"])
        self.database.save()

    def transformations(self, transformation_type = 2, fname_surface=""):

        self.database.create_table("transformation_mapping", self.name_peak_list)
        
        if len(fname_surface) > 0 and type(fname_surface) == str:
            mass_erors.load_surface(fname_surface)

        print
        print "Peak-pair differences to transformations - %s" % (self.file_name_peak_list)
        print "------------------------------------------------------------------"
        perc = 0.0

        columns_insert = ["peaka", "peakb", "peaka_intensity", "peakb_intensity", "ion_peaka", "ion_peakb",
            "transformation", "peak_pair_adducts_diff", "theo_peak_pair_adducts_diff", "ppm_error", "type", "ion_mode"]

        total_iters = ((self.num_peaks*self.num_peaks)/2) + self.num_peaks/2
        iter_done = 0
        
        for item in self.peak_list.peak_pair_diffs():

            if len(fname_surface) > 0 and type(fname_surface) == str:
                mass_error_peak_pair_diff = self.calc.mass_error_peak_pair_diff(item["mean_peakab"], item["peak_pair_diff"], "surface")
            else:
                mass_error_peak_pair_diff = self.calc.abs_error(item["peaka"], self.ppm) + self.calc.abs_error(item["peakb"], self.ppm)
                
            min_tol_peak_pair_diff = item["peak_pair_diff"] - mass_error_peak_pair_diff
            max_tol_peak_pair_diff = item["peak_pair_diff"] + mass_error_peak_pair_diff

            x = ""
            if item["peak_pair_diff"] == 0.0:
                columns_select = [item["peaka"], item["peakb"], item["peaka_intensity"], item["peakb_intensity"],
                "ion_compoundid_1", "ion_compoundid_2", "transformation", "0.0", "0.0", "0.0", "type", "ion_mode"]
                where = "WHERE peak_pair_adducts_diff = 0.0 AND type <= %s and ion_mode = '%s'" % (transformation_type, self.ion_mode)

                self.database.insert_select("transformations_%s" % (self.name_peak_list), columns_insert, columns_select, "unique_transformations", where)
            
            else:
                if item["peaka"] < item["peakb"]:

                    columns_select_1 = [item["peaka"], item["peakb"], item["peaka_intensity"], item["peakb_intensity"],
                    "ion_compoundid_1", "ion_compoundid_2", "transformation", item["peakb"] - item["peaka"], "peak_pair_adducts_diff",
                    "(ABS(peak_pair_adducts_diff)-%s)/abs(peak_pair_adducts_diff*0.000001)" % (abs(item["peaka"] - item["peakb"])), "type", "ion_mode"]

                    where_1 = """WHERE
                    (peak_pair_adducts_diff >= %s AND peak_pair_adducts_diff <= %s)
                    AND type <= %s and ion_mode = '%s'""" % (abs(min_tol_peak_pair_diff), abs(max_tol_peak_pair_diff), transformation_type, self.ion_mode)

                    columns_select_2 = [item["peakb"], item["peaka"], item["peakb_intensity"], item["peaka_intensity"],
                    "ion_compoundid_1", "ion_compoundid_2", "transformation", item["peaka"] - item["peakb"], "peak_pair_adducts_diff",
                    "(ABS(peak_pair_adducts_diff)-%s)/abs(peak_pair_adducts_diff*0.000001)" % (abs(item["peaka"] - item["peakb"])), "type", "ion_mode"]

                    where_2 = """WHERE
                    (peak_pair_adducts_diff <= -%s AND peak_pair_adducts_diff >= -%s)
                    AND type <= %s and ion_mode = '%s'""" % (abs(min_tol_peak_pair_diff), abs(max_tol_peak_pair_diff), transformation_type, self.ion_mode)                 

                    self.database.insert_select("transformations_%s" % (self.name_peak_list), columns_insert, columns_select_1, "unique_transformations", where_1)
                    self.database.insert_select("transformations_%s" % (self.name_peak_list), columns_insert, columns_select_2, "unique_transformations", where_2)

                elif item["peaka"] > item["peakb"]:
                
                    columns_select_1 = [item["peaka"], item["peakb"], item["peaka_intensity"], item["peakb_intensity"],
                    "ion_compoundid_1", "ion_compoundid_2", "transformation", item["peakb"] - item["peaka"], "peak_pair_adducts_diff",
                    "(ABS(peak_pair_adducts_diff)-%s)/abs(peak_pair_adducts_diff*0.000001)" % (abs(item["peaka"] - item["peakb"])), "type", "ion_mode"]

                    where_1 = """WHERE
                    (peak_pair_adducts_diff >= -%s AND peak_pair_adducts_diff <= -%s)
                    AND type <= %s and ion_mode = '%s'""" % (abs(max_tol_peak_pair_diff), abs(min_tol_peak_pair_diff), transformation_type, self.ion_mode)

                    columns_select_2 = [item["peakb"], item["peaka"], item["peakb_intensity"], item["peaka_intensity"],
                    "ion_compoundid_2", "ion_compoundid_1", "transformation", item["peaka"] - item["peakb"], "peak_pair_adducts_diff",
                    "(ABS(peak_pair_adducts_diff)-%s)/abs(peak_pair_adducts_diff*0.000001)" % (abs(item["peaka"] - item["peakb"])), "type", "ion_mode"]

                    where_2 = """WHERE
                    (peak_pair_adducts_diff >= %s AND peak_pair_adducts_diff <= %s)
                    AND type <= %s and ion_mode = '%s'""" % (abs(min_tol_peak_pair_diff), abs(max_tol_peak_pair_diff), transformation_type, self.ion_mode)

                    self.database.insert_select("transformations_%s" % (self.name_peak_list), columns_insert, columns_select_1, "unique_transformations", where_1)
                    self.database.insert_select("transformations_%s" % (self.name_peak_list), columns_insert, columns_select_2, "unique_transformations", where_2)

            iter_done += 1
            if float(iter_done)/total_iters >= perc:
                str_perc = str(perc * 100.0) + "%"
                perc += 0.1
                print str_perc,

        self.database.index("transformations_%s" % (self.name_peak_list), "transformations_%s" % (self.name_peak_list), ["peaka", "transformation", "type", "ion_mode"])
        self.database.save()                     

        print
        print "------------------------------------------------------------------"
        print "DONE"
        print


    def reactant_pairs(self, transformation_type = 2):

        self.database.create_table("reactant_pairs_peaklist", self.name_peak_list)

        peaks_done, perc = 0, 0.0
        print
        print "Assigning reactant pairs - %s" % (self.file_name_peak_list)
        print "------------------------------------------------------------------"

        for mz in self.peak_list.intensities:
            
            self.calc.tolerance_peak(mz, self.ppm)
            where = "where theo_mass_adduct >= %s AND theo_mass_adduct <= %s" % (self.calc.min_tol, self.calc.max_tol)
            compounds = self.database.select("compounds", ["compoundid", "ion"], where)
            if len(compounds) > 0:
                for compound in compounds:
                    self.database.execute_complex("reactant_pairs_temp", [self.name_peak_list, self.name_peak_list,
                    str(compound["compoundid"]), mz, str(compound["ion"]), transformation_type, self.ion_mode])

            peaks_done += 1
            if float(peaks_done)/self.num_peaks >= perc:
                str_perc = str(perc * 100.0) + "%"
                perc += 0.1
                self.database.save()
                print str_perc,

        self.database.save()
        self.database.execute_complex("reactant_pairs", [])
        self.database.save()
        
        print
        print "------------------------------------------------------------------"
        print "DONE"
        print


    def formulae(self, atoms, ions):
        
        print
        print "Calculate formulae - %s" % (self.file_name_peak_list)
        print "------------------------------------------------------------------"

        self.database.create_table("formulae", self.name_peak_list)            
        peaks_done, perc = 0, 0.0

        for mz in self.peak_list.intensities:
            for assignment in self.calc.formulae(mz, self.ppm, atoms, ions):
                self.database.insert("formulae_%s" % (self.name_peak_list),
                {"peak":assignment["mz"], "intensity":0.0, "ion":assignment["ion"], "ppm_error":assignment["ppm_error"], "theoretical_mass":assignment["theoretical_mass"],
                "formula":assignment["formula"], "lewis":assignment["lewis"], "senior":assignment["senior"], "HC":assignment["HC"],
                "NOPSC":assignment["NOPSC"], "ion_mode":self.ion_mode})
            peaks_done += 1
            if float(peaks_done)/self.num_peaks >= perc:
                str_perc = str(perc * 100.0) + "%"
                perc += 0.1
                self.database.save()
                print str_perc,
                        
        self.database.save()
        print
        print "------------------------------------------------------------------"
        print "DONE"
        print

    def formulae_pp(self, atoms, ions):

        import sys
        try:
            import pp
        except:
            print "Please, install Parallel Python: www.parallelpython.com"
            sys.exit()
        
        self.database.create_table("formulae", self.name_peak_list)           

        peaks_done, perc = 0, 0.0
        jobs = []
        ppservers = ()
        job_server = pp.Server(ppservers=ppservers)
        print
        print "Calculate formulae - %s" % (self.file_name_peak_list)
        print "Starting Parallel Python with", job_server.get_ncpus(), "workers"
        print "------------------------------------------------------------------"

        def temp_pp(mass, ppm, atoms, ions):
            calc = Calculate()
            sql_inserts = []
            assignments = calc.formulae(mass, ppm, atoms, ions)
            for assignment in assignments:
                sql_inserts.append(assignment)
            return sql_inserts

        for mz in self.peak_list.intensities:
            jobs.append(job_server.submit(temp_pp, (mz, self.ppm, atoms, ions), (Calculate, Formula), ("re", "copy")))
        for job in jobs:
            job = job()
            for assignment in job:
                self.database.insert("formulae_%s" % (self.name_peak_list),
                {"peak":assignment["mz"], "intensity":self.peak_list.intensity(assignment["mz"]), "ion":assignment["ion"], "ppm_error":assignment["ppm_error"], "theoretical_mass":assignment["theoretical_mass"],
                "formula":assignment["formula"], "lewis":assignment["lewis"], "senior":assignment["senior"], "HC":assignment["HC"],
                "NOPSC":assignment["NOPSC"], "ion_mode":self.ion_mode})
            peaks_done += 1
            if float(peaks_done)/self.num_peaks >= perc:
                str_perc = str(perc * 100.0) + "%"
                perc += 0.1
                self.database.save()
                print str_perc,            
        self.database.save()
        print
        print "------------------------------------------------------------------"

    def peak_patterns(self, ions, ppm=1.0, min_adducts=2, rules = {}, pattern_type = "adducts"):
        from copy import deepcopy

        table_flag = self.database.select("sqlite_master", ["name"], "WHERE type='table' and name = 'peak_patterns_%s'" % (self.name_peak_list))

        if len(table_flag) == 0:
            self.database.create_table("peak_patterns", self.name_peak_list)       

        isotopes_count = self.database.count("peak_patterns_%s" % (self.name_peak_list), "peaka_label",
            "where pattern_type == 'isotopes'")
        adducts_count = self.database.count("peak_patterns_%s" % (self.name_peak_list), "peaka_label",
            "where pattern_type == 'adducts'")

        if isotopes_count > 0 and pattern_type == "isotopes":
            self.database.delete("peak_patterns_%s" % (self.name_peak_list),
            "where pattern_type == 'isotopes'")
            self.database.save()

        if adducts_count > 0 and pattern_type == "adducts":
            self.database.delete("peak_patterns_%s" % (self.name_peak_list),
            "where pattern_type == 'adducts'")
            self.database.save()

        peaks_done, perc = 0, 0.0
        print
        print "Assigning peak patterns - %s" % (self.file_name_peak_list)
        print "------------------------------------------------------------------"

        peak_id = 0
        lib_masses = {}
        sort_ids = []
        pattern_id = 0
     
        for ion in ions.lib:
            for mz in self.peak_list.intensities:

                peak_id += 1
                peak_min_ion = mz - ions.lib[ion]

                if peak_min_ion > 0:

                    lib_masses[peak_id] = [peak_min_ion, mz, ion, self.peak_list.intensity(mz), ions.lib[ion]]

        #### sort peak list from high to low based on original peak
                    
                    peak_error = self.calc.abs_error(mz, self.ppm)
                    min_tol = peak_min_ion - peak_error
                    max_tol = peak_min_ion + peak_error

                    sort_ids.append([peak_id, min_tol+max_tol])
               
        sort_ids.sort(lambda x, y: cmp(y[1],x[1]))

        for x in range(len(sort_ids)):
            sort_ids[x] = sort_ids[x][0]


        ############################################################

        copy_sort_ids = deepcopy(sort_ids)
        
        for peak_id in sort_ids:

            peaks_done += 1
            
            if peak_id in copy_sort_ids: # check if not already assigned - avoiding double records
                
                peak_error = self.calc.abs_error(lib_masses[peak_id][1], self.ppm)
                min_tol = float(lib_masses[peak_id][0]) - peak_error
                max_tol = float(lib_masses[peak_id][0]) + peak_error

                pattern = []
                ions_pattern = []
                for comp_peak_id in copy_sort_ids:
                    
                    if lib_masses[comp_peak_id][0] >= min_tol and lib_masses[comp_peak_id][0] <= max_tol:
                        pattern.append(comp_peak_id)
                        ions_pattern.append(lib_masses[comp_peak_id][2])

                if len(pattern) >= min_adducts: # check number of peaks in pattern

                    pattern_id += 1
                    for i in pattern:
                        copy_sort_ids.remove(i) # remove peak id to avoid duplications
                        for j in pattern:
                            if lib_masses[i][1] - lib_masses[j][1] > 0.0:
                                rules_out = 1
                                carbons = 0
                                if len(set(ions_pattern)) == len(ions_pattern): # check for double ions in pattern
                                    
                                    if len(rules) > 0:
                                        for rule in rules: # check rules defined by user
                                            if rule[rule.keys()[0]] == ions_pattern.count(rule.keys()[0]) and rule[rule.keys()[1]] == ions_pattern.count(rule.keys()[1]):
                                                rules_out = 0
                                                break
                                else:
                                    rules_out = 0
                                
                                if pattern_type == "isotopes" and rules_out != 0: # check inensities isotopes if isotope pattern
                                    max_intensity = (-0.0213 * lib_masses[i][1]) + 100.24
                                    min_intensity = (-0.0768 * lib_masses[i][1]) + 98.693
                                    check_inten = 100 - (lib_masses[i][3]/(lib_masses[i][3]+lib_masses[j][3]))*100

                                    if check_inten >= min_intensity and check_inten <= max_intensity :
                                        rules_out = 1

                                    else:
                                        rules_out = 0

                                    carbons = (100*float(lib_masses[i][3]))/(1.1*float(lib_masses[j][3]))
                                self.database.insert("peak_patterns_%s" %(self.name_peak_list),
                                    {"id":pattern_id, "peak_pair_diff":lib_masses[i][1] - lib_masses[j][1],
                                    "peaka":lib_masses[j][1], "peakb":lib_masses[i][1],
                                    "peaka_intensity":lib_masses[j][3], "peakb_intensity":lib_masses[i][3],
                                    "peaka_label":lib_masses[j][2], "peakb_label":lib_masses[i][2],
                                    "pattern_type":pattern_type, "num_peaks":len(pattern),
                                    "carbons":carbons, "rules":rules_out, "ion_mode":self.ion_mode}) 

                                     
            if float(peaks_done)/len(sort_ids) >= perc:
                print str(perc*100)+"%",
                perc += 0.1
                self.database.save()
        
        self.database.save()

        try:

            self.database.index("id_%s_peak_patterns" % (self.name_peak_list),
                "%s_peak_patterns" % (self.name_peak_list),["peaka", "peakb"])

            self.database.index("peakab_%s_peak_patterns" % (self.name_peak_list),
                "%s_peak_patterns" % (self.name_peak_list),["id"])

            self.database.index("%s_peak_patterns" % (self.name_peak_list),
                "%s_peak_patterns" % (self.name_peak_list),
                ["rules", "peaka_label", "peakb_label", "num_peaks", "peaka", "peakb"])
            
        except:
            pass

        print
        print "------------------------------------------------------------------"
        print "DONE"
        print
