#!/usr/local/bin/python

from MI_Pack import Atoms
from MI_Pack import Ions
from MI_Pack import KEGG
from MI_Pack import Identification

# MI-Pack: Increased confidence of metabolite identification in mass spectra by integrating accurate
# masses and metabolic pathways
# Ralf J.M. Weber (1) and Mark R. Viant (1,2) - rxw744@bham.ac.uk, m.viant@bham.ac.uk
# (1) Centre for Systems Biology, University of Birmingham, Edgbaston, Birmingham, B15 2TT, UK 
# (2) School of Biosciences, University of Birmingham, Edgbaston, Birmingham, B15 2TT, UK
# 
# Examples Metabolite Identification Package (MI-Pack) 
#
# See manual for more examples

atoms = Atoms() #Create standard set of atoms
ions = Ions("POS") #Create standard set of "positive" ion forms (e.g. H+, Na+, K(39)+)
#ions = Ions("NEG") #Create standard set of "negative" ion forms

print atoms
print ions

# Download KEGG database files
# - compound -> compound records
# - reaction_mapformula.lst -> reactant pair records
download = raw_input("Download KEGG database files  y/n: ")
if download == "Y" or download == "y":

    import urllib # This module provides a high-level interface for fetching data across the World Wide Web
    urllib.urlretrieve("ftp://ftp.genome.jp/pub/kegg/ligand/reaction/reaction_mapformula.lst", "reaction_mapformula.lst")
    urllib.urlretrieve("ftp://ftp.genome.jp/pub/kegg/ligand/compound/compound", "compound")

# Create database -> MI_DB_POS.sqlite
create_db = raw_input("Create Metabolite Identification Database -> MI_DB_POS.sqlite  y/n: ")
if create_db == "Y" or create_db == "y":

    KEGG.Compounds("compound", atoms, ions, "POS")
    KEGG.ReactantPairs("reaction_mapformula.lst", atoms, ions, "POS", 1)
    # 1 -> Parse direct reactant pairs/transformations
    # 2 -> Parse direct AND indirect reactant pairs/transformations

#Metabolite names, Transformations and reactant pairs -> db_example_ID.sqlite
ID = Identification("peaklist_TCA_cycle.txt", "db_example_ID", "POS", 0.0) # Create dataset for identification - 0.0 ppm error
ID.single_peak_search() # Single-peak search

ID.transformations(1)   #1 -> Assign direct transformations to peak-pair differences
                        #2 -> Assign direct OR/AND indirect transformations to peak-pair differences

ID.reactant_pairs(1)    #1 -> Assign direct transformations to peak-pair differences
                        #2 -> Assign direct OR/AND indirect transformations to peak-pair differences

#A zero ppm error was used for the identification approaches as the list comprises of
#theoretical masses, which are identical to the mass values in the MI-DB.

#Emperical formulae -> db_example_formulae.sqlite
ID_ef = Identification("peaklistA.txt", "db_example_formulae", "POS", 1.0) # Create dataset for identification - 1.0 ppm error
ID_ef.formulae(atoms, ions)

#Peak patterns -> db_example_peak_patterns.sqlite
ID_pp = Identification("peaklistB.txt", "db_example_peak_patterns", "POS", 1.0) # Create dataset for identification - 1.0 ppm error
ID_pp.peak_patterns(ions)









